/*
 * Copyright (C) 2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */
#include "gcc-compat.h"
#include "para.h"
#include "close_on_fork.h"


#include "grab_client.cmdline.h"
#include "list.h"
#include "filter.h"
#include "grab_client.h"
#include "audiod.h"
#include "error.h"


static const char *gc_modes[] = {
	[GRAB_PEDANTIC] = "pedantic",
	[GRAB_SLOPPY] = "sloppy",
	[GRAB_AGGRESSIVE] = "aggressive",
	NULL
};

const char *audio_formats[] = {AUDIO_FORMAT_ARRAY};

LIST_HEAD(inactive_grab_client_list);


/* FIXME: -> audiod */
int audio_format_num(const char *name)
{
	int i;

	for (i = 0; audio_formats[i]; i++)
		if (!strcmp(audio_formats[i], name))
			return i;
	return -E_GC_AUDIO_FORMAT;
}


static int max_num_filters(void)
{
	int i, ret = 0;
	for (i = 0; audio_formats[i]; i++) {
		PARA_INFO_LOG("%s filter chain length: %d\n", audio_formats[i],
			 num_filters(i));
		ret = MAX(ret, num_filters(i));
	}
	PARA_INFO_LOG("maximal filter chain length: %d\n", ret);
	return ret;
}

static int gc_write(char *buf, size_t len, struct filter_callback *fcb)
{
	struct grab_client *gc = fcb->data;
	struct timeval tv = {0, 100};
	int ret;

//	PARA_INFO_LOG("writing %d bytes to fd %d\n", len, gc->fd);
	fd_set wfds;
	do {
		FD_ZERO(&wfds);
		FD_SET(gc->fd, &wfds);
		ret = select(gc->fd + 1, NULL, &wfds, NULL, &tv);
	} while (ret == EAGAIN || ret == EINTR);
	if (ret != 1) {
		if (gc->mode == GRAB_PEDANTIC)
			return -E_PEDANTIC_GRAB;
		if (gc->mode == GRAB_SLOPPY)
			return 1;
	}
rewrite:
	ret = write(gc->fd, buf, len);
	if (ret < 0) {
		ret = -E_GC_WRITE;
		gc->error = E_GC_WRITE;
	} else {
		if (ret != len) {
			if (gc->mode == GRAB_PEDANTIC)
				return -E_PEDANTIC_GRAB;
			if (gc->mode == GRAB_AGGRESSIVE) {
				len -= ret;
				memmove(buf, buf + ret, len);
				goto rewrite;
			}
		}
	}
	return ret;
}

static int check_gc_args(struct grab_client *gc)
{
	int i;
	struct grab_client_args_info *conf = gc->conf;

	PARA_INFO_LOG("filter_num: %d\n", gc->conf->filter_num_arg);
	for (i = 0; gc_modes[i]; i++)
		if (!strcmp(conf->mode_arg, gc_modes[i]))
			break;
	if (!gc_modes[i])
		return -E_INVALID_GRAB_MODE;
	gc->mode = i;
	if (conf->audio_format_given) {
		gc->audio_format_num = audio_format_num(conf->audio_format_arg);
		if (gc->audio_format_num < 0)
			return gc->audio_format_num;
	}
	if (conf->slot_arg > MAX_STREAM_SLOTS)
		return -E_BAD_GC_SLOT;
	if (conf->filter_num_arg <= 0)
		return -E_BAD_GC_FILTER_NUM;
	if (conf->audio_format_given) {
		if (num_filters(gc->audio_format_num) < conf->filter_num_arg)
			return -E_BAD_GC_FILTER_NUM;
	} else
		if (conf->filter_num_arg > max_num_filters())
			return -E_BAD_GC_FILTER_NUM;

	return 1;
}

static void add_inactive_gc(struct grab_client *gc)
{
	PARA_INFO_LOG("adding grab client %p (fd %d) to inactive list\n",
		gc, gc->fd);
	list_add(&gc->node, &inactive_grab_client_list);
}

static void gc_free(struct grab_client *gc)
{
	int i;

	for (i = 0; i < gc->argc; i++)
		free(gc->argv[i]);
	free(gc->argv);
	free(gc->conf);
	free(gc);

}

static void gc_close(struct filter_callback *fcb)
{
	
	struct grab_client *gc = fcb->data;
	if (gc->conf->one_shot_given || gc->error) {
		PARA_INFO_LOG("closing fd %d (grab client %p)\n", gc->fd, gc);
		del_close_on_fork_list(gc->fd);
		close(gc->fd);
		gc_free(gc);
		/* close on fork ?*/
		return;
	}
	add_inactive_gc(gc);
}

void activate_grab_client(struct grab_client *gc, struct filter_node *fn)
{
	PARA_INFO_LOG("activating %p (fd %d, filter node: %p)\n", gc, gc->fd, fn);
	list_del(&gc->node);
	list_add(&gc->fcb.node, &fn->callbacks);
}

void activate_inactive_grab_clients(int slot, int audio_format_num,
		struct list_head *filter_list)
{
	struct grab_client *gc, *tmp;
	int i;
	struct filter_node *fn;

	list_for_each_entry_safe(gc, tmp, &inactive_grab_client_list, node) {
//		PARA_INFO_LOG("checking inactive grab client %p\n", gc);
		if (gc->conf->slot_arg >= 0 && gc->conf->slot_arg != slot)
			continue;
		if (gc->audio_format_num >= 0 && gc->audio_format_num !=
				audio_format_num)
			continue;
		if (gc->conf->filter_num_arg >= 0 &&
				num_filters(gc->audio_format_num)
				< gc->conf->filter_num_arg)
			continue;
		i = 1;
		list_for_each_entry(fn, filter_list, node) {
			if (gc->conf->filter_num_arg <= 0
				|| i == gc->conf->filter_num_arg)
				break;
			i++;
		}
		activate_grab_client(gc, fn);
	}
}

__malloc struct grab_client *grab_client_new(int fd, int argc, char **argv, int *err)
{
	int i, ret;
	struct grab_client *gc = para_calloc(sizeof(struct grab_client));

	gc->conf = para_calloc(sizeof(struct grab_client_args_info));
	gc->argc = argc;
	gc->argv = para_calloc((argc + 1) * sizeof(char *));

	for (i = 0; i < argc; i++) {
		gc->argv[i] = para_strdup(argv[i]);
		PARA_INFO_LOG("argc: %d, argv[%d]: %s\n", argc, i, gc->argv[i]);
	}
	PARA_INFO_LOG("argv[%d]: %s\n", argc, gc->argv[argc]);
	ret = grab_client_cmdline_parser(gc->argc, gc->argv , gc->conf);
	*err = -E_GC_SYNTAX;
	if (ret)
		goto err_out;
	*err = -E_GC_HELP_GIVEN;
	if (gc->conf->help_given)
		goto err_out;
	*err = check_gc_args(gc);
	if (*err < 0)
		goto err_out;
	if (gc->conf->input_grab_given) {
		gc->fcb.input_cb = gc_write;
		gc->fcb.output_cb = NULL;
	} else {
		gc->fcb.output_cb = gc_write;
		gc->fcb.input_cb = NULL;
	}
	gc->fd = fd;
	gc->fcb.close = gc_close;
	gc->fcb.data = gc;
	add_inactive_gc(gc);
	return gc;
err_out:
	for (i = 0; i < argc; i++)
		free(gc->argv[i]);
	free(gc->argv);
	free(gc->conf);
	free(gc);
	return NULL;
}

void init_grabbing()
{
	PARA_INFO_LOG("%s", "grab init\n");
	INIT_LIST_HEAD(&inactive_grab_client_list);
}

